#!/usr/bin/env cpython
#
# pyLOM - Python Low Order Modeling.
#
# Python interface for DMD.
#
# Last rev: 30/09/2021
from __future__ import print_function, division

cimport cython
cimport numpy as np

import numpy as np

#from libc.complex  cimport creal, cimag
cdef extern from "<complex.h>" nogil:
	float  complex I
	# Decomposing complex values
	float cimagf(float complex z)
	float crealf(float complex z)
	double cimag(double complex z)
	double creal(double complex z)
cdef double complex J = 1j
from libc.stdlib     cimport malloc, free
from libc.string     cimport memcpy, memset
from libc.math       cimport sqrt, log, atan2
from ..vmmath.cfuncs cimport real, real_complex
from ..vmmath.cfuncs cimport c_stranspose, c_smatmul, c_smatmulp, c_svecmat, c_stemporal_mean, c_ssubtract_mean, c_stsqr_svd, c_ssvd, c_scompute_truncation_residual, c_scompute_truncation
from ..vmmath.cfuncs cimport c_dtranspose, c_dmatmul, c_dmatmulp, c_dvecmat, c_dtemporal_mean, c_dsubtract_mean, c_dtsqr_svd, c_dsvd, c_dcompute_truncation_residual, c_dcompute_truncation
from ..vmmath.cfuncs cimport c_cmatmult, c_cvecmat, c_cinverse, c_ccholesky, c_ceigen, c_cvandermonde, c_cvandermonde_time, c_csort
from ..vmmath.cfuncs cimport c_zmatmult, c_zvecmat, c_zinverse, c_zcholesky, c_zeigen, c_zvandermonde, c_zvandermonde_time, c_zsort

from ..utils.cr       import cr, cr_start, cr_stop
from ..utils.errors   import raiseError


## DMD run method
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _srun(float[:,:] X, float r, int remove_mean):
	'''
	Run DMD analysis of a matrix X.

	Inputs:
		- X[ndims*nmesh,n_temp_snapshots]: data matrix
		- remove_mean:                     whether or not to remove the mean flow
		- r:                               maximum truncation residual

	Returns:
		- Phi:      DMD Modes
		- muReal:   Real part of the eigenvalues
		- muImag:   Imaginary part of the eigenvalues
		- b:        Amplitude of the DMD modes
		- Variables needed to reconstruct flow
	'''
	# Variables
	cdef int m = X.shape[0], n = X.shape[1], mn = min(m,n-1), retval
	cdef float *X_mean
	cdef float *Y
	cdef int iaux, icol, irow
	# Output arrays:
	# Allocate memory
	Y  = <float*>malloc(m*n*sizeof(float))

	# Remove mean if required
	if remove_mean:
		cr_start('DMD.temporal_mean',0)
		X_mean = <float*>malloc(m*sizeof(float))
		# Compute temporal mean
		c_stemporal_mean(X_mean,&X[0,0],m,n)
		# Compute substract temporal mean
		c_ssubtract_mean(Y,&X[0,0],X_mean,m,n)
		free(X_mean)
		cr_stop('DMD.temporal_mean',0)
	else:
		memcpy(Y,&X[0,0],m*n*sizeof(float))

	# Get the first N-1 snapshots: Y1 = Y[:,:-1]
	cr_start('DMD.split_snapshots', 0)
	cdef float *Y1
	cdef float *Y2
	Y1 = <float*>malloc(m*(n-1)*sizeof(float))
	Y2 = <float*>malloc(m*(n-1)*sizeof(float))
	for irow in range(m):
		for icol in range(n-1):
			Y1[irow*(n-1) + icol] = Y[irow*n + icol]
			Y2[irow*(n-1) + icol] = Y[irow*n + icol + 1]
	free(Y)
	cr_stop('DMD.split_snapshots', 0)

	# Compute SVD
	cr_start('DMD.SVD',0)
	cdef float *U
	cdef float *S
	cdef float *V
	U  = <float*>malloc(m*mn*sizeof(float))
	S  = <float*>malloc(mn*sizeof(float))
	V  = <float*>malloc((n-1)*mn*sizeof(float))
	retval = c_stsqr_svd(U, S, V, Y1, m, mn)
	cr_stop('DMD.SVD',0)
	if not retval == 0: raiseError('Problems computing SVD!')
	free(Y1)

	# Truncate
	cr_start('DMD.truncate',0)
	cdef int nr
	cdef float *Ur
	cdef float *Sr
	cdef float *Vr

	nr = int(r) if r > 1 else c_scompute_truncation_residual(S,r,n-1)
	Ur = <float*>malloc(m*nr*sizeof(float))
	Sr = <float*>malloc(nr*sizeof(float))
	Vr = <float*>malloc(nr*mn*sizeof(float))
	c_scompute_truncation(Ur,Sr,Vr,U,S,V,m,n-1,n-1,nr)
	
	free(U)
	free(V)
	free(S)
	cr_stop('DMD.truncate',0)

	# Project Jacobian of the snapshots into the POD basis
	cr_start('DMD.linear_mapping',0)
	cdef float *aux1
	cdef float *aux2
	cdef float *aux3
	cdef float *Atilde
	cdef float *Urt
	aux1   = <float*>malloc(nr*(n-1)*sizeof(float))
	aux2   = <float*>malloc(nr*(n-1)*sizeof(float))
	aux3   = <float*>malloc(nr*sizeof(float))
	Atilde = <float*>malloc(nr*nr*sizeof(float))
	Urt    = <float*>malloc(nr*m*sizeof(float))
	c_stranspose(Ur, Urt, m, nr)
	c_smatmulp(aux1, Urt, Y2, nr, n-1, m)
	for icol in range(n-1):
		for irow in range(nr):
			aux2[icol*nr + irow] = Vr[irow*(n-1) + icol]/Sr[irow]
	c_smatmul(Atilde, aux1, aux2, nr, nr, n-1)
	free(aux1)
	free(aux3)
	free(Urt)
	cr_stop('DMD.linear_mapping',0)

	# Compute eigenmodes
	cdef float *auxmuReal
	cdef float *auxmuImag
	cdef np.complex64_t *w
	auxmuReal = <float*>malloc(nr*sizeof(float))
	auxmuImag = <float*>malloc(nr*sizeof(float))
	w         = <np.complex64_t*>malloc(nr*nr*sizeof(np.complex64_t))
	cr_start('DMD.eigendecomposition',0)
	retval = c_ceigen(auxmuReal,auxmuImag,w,Atilde,nr,nr)
	cr_stop('DMD.eigendecomposition',0)
	free(Atilde)

	# Computation of DMD modes
	cr_start('DMD.modes',0)
	cdef np.complex64_t *auxPhi
	cdef np.complex64_t *aux1C
	cdef np.complex64_t *aux2C
	auxPhi = <np.complex64_t*>malloc(m*nr*sizeof(np.complex64_t))
	aux1C  = <np.complex64_t*>malloc(nr*sizeof(np.complex64_t))
	aux2C  = <np.complex64_t*>malloc(nr*sizeof(np.complex64_t))
	for iaux in range(m):
		for icol in range(nr):
			aux1C[icol] = 0 + 0*I
			for irow in range(n-1):
				aux1C[icol] += Y2[iaux*(n-1) + irow]*aux2[irow*nr + icol]
		c_cmatmult(aux2C, aux1C, w, 1, nr, nr, 'N', 'N')
		memcpy(&auxPhi[iaux*nr], aux2C, nr*sizeof(np.complex64_t))
	free(aux2)
	free(Y2)
	cdef float a
	cdef float b
	cdef float c
	cdef float d
	cdef float div
	for icol in range(nr):
		c = auxmuReal[icol]
		d = auxmuImag[icol]
		div = c*c + d*d
		for iaux in range(m):
			a = crealf(auxPhi[iaux*nr + icol])
			b = cimagf(auxPhi[iaux*nr + icol])
			auxPhi[iaux*nr + icol] = (a*c + b*d)/div + (b*c - a*d)/div*I
	cr_stop('DMD.modes',0)

	# Amplitudes according to: Jovanovic et. al. 2014 DOI: 10.1063
	cdef np.complex64_t *auxbJov
	cdef np.complex64_t *aux3C
	cdef np.complex64_t *Vand
	cdef np.complex64_t *P
	cdef np.complex64_t *Pinv
	cdef np.complex64_t *q

	auxbJov = <np.complex64_t*>malloc(nr*sizeof(np.complex64_t))
	aux3C   = <np.complex64_t*>malloc(nr*nr*sizeof(np.complex64_t))
	aux4C   = <np.complex64_t*>malloc(nr*nr*sizeof(np.complex64_t))
	Vand    = <np.complex64_t*>malloc((nr*(n-1))*sizeof(np.complex64_t))
	P       = <np.complex64_t*>malloc(nr*nr*sizeof(np.complex64_t))
	Pinv    = <np.complex64_t*>malloc(nr*nr*sizeof(np.complex64_t))
	q       = <np.complex64_t*>malloc(nr*sizeof(np.complex64_t))

	cr_start('DMD.amplitudes', 0)
	c_cvandermonde(Vand, auxmuReal, auxmuImag, nr, n-1)
	c_cmatmult(aux3C, w, w, nr, nr, nr, 'C', 'N')
	c_cmatmult(aux4C, Vand, Vand, nr, nr, n-1, 'N', 'C')

	for irow in range(nr):
		for icol in range(nr): # Loop on the columns of the Vandermonde matrix
			P[irow*nr + icol]  = crealf(aux3C[irow*nr + icol])*crealf(aux4C[irow*nr + icol])
			P[irow*nr + icol] += -crealf(aux3C[irow*nr + icol])*cimagf(aux4C[irow*nr + icol])*I 
			P[irow*nr + icol] += cimagf(aux3C[irow*nr + icol])*crealf(aux4C[irow*nr + icol])*I
			P[irow*nr + icol] += cimagf(aux3C[irow*nr + icol])*cimagf(aux4C[irow*nr + icol])
	retval = c_ccholesky(P, nr)
	if not retval == 0: raiseError('Problems computing Cholesky factorization!')

	for iaux in range(nr):
		for irow in range(nr):
			aux1C[irow] = 0 + 0*I
			for icol in range(n-1):# casting Vr to a complex, at the same time, it is multipilied per S and Vand
				aux1C[irow] += Sr[irow]*Vr[irow*(n-1) + icol]*(crealf(Vand[iaux*(n-1) + icol])+cimagf(Vand[iaux*(n-1) + icol])*I)
			aux2C[irow] = w[irow*nr + iaux]
		c_cmatmult(&q[iaux], aux1C, aux2C, 1, 1, nr, 'N', 'N')

	memcpy(Pinv, P, nr*nr*sizeof(np.complex64_t))
	cdef int ii
	cdef int jj
	for ii in range(nr):
		q[ii] = crealf(q[ii]) - cimagf(q[ii])*I
		for jj in range(nr - ii):
			P[ii*nr + ii+jj]   = crealf(P[(ii+jj)*nr + ii])  - cimagf(P[(ii+jj)*nr + ii])*I
			P[(ii+jj)*nr + ii] = crealf(Pinv[ii*nr + ii+jj]) - cimagf(Pinv[ii*nr + ii+jj])*I

	retval = c_cinverse(Pinv, nr, 'L')
	if not retval == 0: raiseError('Problems computing the Inverse!')

	c_cmatmult(aux1C, Pinv, q, nr, 1, nr, 'N', 'N')

	retval = c_cinverse(P, nr, 'U')
	if not retval == 0: raiseError('Problems computing the Inverse!')

	c_cmatmult(auxbJov, P, aux1C, nr, 1, nr, 'N', 'N')
	cr_stop('DMD.amplitudes',0)

	# Free allocated arrays before reordering
	free(Ur)
	free(Sr)
	free(Vr)
	free(aux1C)
	free(aux2C)
	free(aux3C)
	free(aux4C)
	free(w)
	free(Vand)
	free(q)
	free(P)
	free(Pinv)

	# Order modes and eigenvalues according to its amplitude
	cdef int *auxOrd
	auxOrd = <int*>malloc(nr*sizeof(int))
	cdef np.ndarray[np.float32_t,ndim=1] muReal   = np.zeros((nr),dtype=np.float)
	cdef np.ndarray[np.float32_t,ndim=1] muImag   = np.zeros((nr),dtype=np.float)
	cdef np.ndarray[np.complex64_t,ndim=2] Phi  = np.zeros((m,nr),order='C',dtype=np.complex64)
	cdef np.ndarray[np.complex64_t,ndim=1] bJov = np.zeros((nr,),dtype=np.complex64)

	cr_start('DMD.qsort', 0)
	c_csort(auxbJov, auxOrd, nr)
	cr_stop('DMD.qsort', 0)
	cr_start('DMD.sort', 0)
	for ii in range(nr):
		muReal[nr-(auxOrd[ii]+1)] = auxmuReal[ii]
		muImag[nr-(auxOrd[ii]+1)] = auxmuImag[ii]
		bJov[nr-(auxOrd[ii]+1)]   = auxbJov[ii]
		for jj in range(m):
			Phi[jj,nr-(auxOrd[ii]+1)]  = auxPhi[jj*nr + ii]
	cr_stop('DMD.sort', 0)

	# Free the variables that had to be ordered
	free(auxmuReal)
	free(auxmuImag)
	free(auxbJov)
	free(auxPhi)
	free(auxOrd)

	# Ensure that all conjugate modes are in the same order
	cr_start('DMD.conjugate', 0)
	cdef bint p = 0
	cdef float iimag
	for ii in range(nr):
		if p == 1:
			p = 0
			continue
		iimag = muImag[ii]
		if iimag < 0:
			muImag[ii]   =  muImag[ii+1]
			muImag[ii+1] = -muImag[ii]
			bJov[ii]     = crealf(bJov[ii])   + cimagf(bJov[ii+1])*I
			bJov[ii+1]   = crealf(bJov[ii+1]) - cimagf(bJov[ii])*I
			for jj in range(m):
				Phi[jj,ii]   = crealf(Phi[jj,ii])   + cimagf(Phi[jj,ii+1])*I
				Phi[jj,ii+1] = crealf(Phi[jj,ii+1]) - cimagf(Phi[jj,ii+1])*I
			p = 1
			continue
		if iimag > 0:
			p = 1
			continue
	cr_stop('DMD.conjugate', 0)
	
	# Return
	return muReal, muImag, Phi, bJov

@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _drun(double[:,:] X, double r, int remove_mean):
	'''
	Run DMD analysis of a matrix X.

	Inputs:
		- X[ndims*nmesh,n_temp_snapshots]: data matrix
		- remove_mean:                     whether or not to remove the mean flow
		- r:                               maximum truncation residual

	Returns:
		- Phi:      DMD Modes
		- muReal:   Real part of the eigenvalues
		- muImag:   Imaginary part of the eigenvalues
		- b:        Amplitude of the DMD modes
		- Variables needed to reconstruct flow
	'''
	# Variables
	cdef int m = X.shape[0], n = X.shape[1], mn = min(m,n-1), retval
	cdef double *X_mean
	cdef double *Y
	cdef int iaux, icol, irow
	# Output arrays:
	# Allocate memory
	Y  = <double*>malloc(m*n*sizeof(double))

	# Remove mean if required
	if remove_mean:
		cr_start('DMD.temporal_mean',0)
		X_mean = <double*>malloc(m*sizeof(double))
		# Compute temporal mean
		c_dtemporal_mean(X_mean,&X[0,0],m,n)
		# Compute substract temporal mean
		c_dsubtract_mean(Y,&X[0,0],X_mean,m,n)
		free(X_mean)
		cr_stop('DMD.temporal_mean',0)
	else:
		memcpy(Y,&X[0,0],m*n*sizeof(double))

	#Get the first N-1 snapshots: Y1 = Y[:,:-1]
	cr_start('DMD.split_snapshots', 0)
	cdef double *Y1
	cdef double *Y2
	Y1 = <double*>malloc(m*(n-1)*sizeof(double))
	Y2 = <double*>malloc(m*(n-1)*sizeof(double))
	for irow in range(m):
		for icol in range(n-1):
			Y1[irow*(n-1) + icol] = Y[irow*n + icol]
			Y2[irow*(n-1) + icol] = Y[irow*n + icol + 1]
	free(Y)
	cr_stop('DMD.split_snapshots', 0)

	# Compute SVD
	cr_start('DMD.SVD',0)
	cdef double *U
	cdef double *S
	cdef double *V
	U  = <double*>malloc(m*mn*sizeof(double))
	S  = <double*>malloc(mn*sizeof(double))
	V  = <double*>malloc((n-1)*mn*sizeof(double))
	retval = c_dtsqr_svd(U, S, V, Y1, m, mn)
	cr_stop('DMD.SVD',0)
	if not retval == 0: raiseError('Problems computing SVD!')
	free(Y1)

	# Truncate
	cr_start('DMD.truncate',0)
	cdef int nr
	cdef double *Ur
	cdef double *Sr
	cdef double *Vr

	nr = int(r) if r > 1 else c_dcompute_truncation_residual(S,r,n-1)
	Ur = <double*>malloc(m*nr*sizeof(double))
	Sr = <double*>malloc(nr*sizeof(double))
	Vr = <double*>malloc(nr*mn*sizeof(double))
	c_dcompute_truncation(Ur,Sr,Vr,U,S,V,m,n-1,n-1,nr)
	
	free(U)
	free(V)
	free(S)
	cr_stop('DMD.truncate',0)

	# Project Jacobian of the snapshots into the POD basis
	cr_start('DMD.linear_mapping',0)
	cdef double *aux1
	cdef double *aux2
	cdef double *aux3
	cdef double *Atilde
	cdef double *Urt
	aux1   = <double*>malloc(nr*(n-1)*sizeof(double))
	aux2   = <double*>malloc(nr*(n-1)*sizeof(double))
	aux3   = <double*>malloc(nr*sizeof(double))
	Atilde = <double*>malloc(nr*nr*sizeof(double))
	Urt    = <double*>malloc(nr*m*sizeof(double))
	c_dtranspose(Ur, Urt, m, nr)
	c_dmatmulp(aux1, Urt, Y2, nr, n-1, m)
	for icol in range(n-1):
		for irow in range(nr):
			aux2[icol*nr + irow] = Vr[irow*(n-1) + icol]/Sr[irow]
	c_dmatmul(Atilde, aux1, aux2, nr, nr, n-1)
	free(aux1)
	free(aux3)
	free(Urt)
	cr_stop('DMD.linear_mapping',0)

	# Compute eigenmodes
	cdef double *auxmuReal
	cdef double *auxmuImag
	cdef np.complex128_t *w
	auxmuReal = <double*>malloc(nr*sizeof(double))
	auxmuImag = <double*>malloc(nr*sizeof(double))
	w         = <np.complex128_t*>malloc(nr*nr*sizeof(np.complex128_t))
	cr_start('DMD.eigendecomposition',0)
	retval = c_zeigen(auxmuReal,auxmuImag,w,Atilde,nr,nr)
	cr_stop('DMD.eigendecomposition',0)
	free(Atilde)

	# Computation of DMD modes
	cr_start('DMD.modes',0)
	cdef np.complex128_t *auxPhi
	cdef np.complex128_t *aux1C
	cdef np.complex128_t *aux2C
	auxPhi = <np.complex128_t*>malloc(m*nr*sizeof(np.complex128_t))
	aux1C  = <np.complex128_t*>malloc(nr*sizeof(np.complex128_t))
	aux2C  = <np.complex128_t*>malloc(nr*sizeof(np.complex128_t))
	for iaux in range(m):
		for icol in range(nr):
			aux1C[icol] = 0 + 0*J
			for irow in range(n-1):
				aux1C[icol] += Y2[iaux*(n-1) + irow]*aux2[irow*nr + icol]
		c_zmatmult(aux2C, aux1C, w, 1, nr, nr, 'N', 'N')
		memcpy(&auxPhi[iaux*nr], aux2C, nr*sizeof(np.complex128_t))
	free(aux2)
	free(Y2)
	cdef double a
	cdef double b
	cdef double c
	cdef double d
	cdef double div
	for icol in range(nr):
		c = auxmuReal[icol]
		d = auxmuImag[icol]
		div = c*c + d*d
		for iaux in range(m):
			a = creal(auxPhi[iaux*nr + icol])
			b = cimag(auxPhi[iaux*nr + icol])
			auxPhi[iaux*nr + icol] = (a*c + b*d)/div + (b*c - a*d)/div*J
	cr_stop('DMD.modes',0)

	# Amplitudes according to: Jovanovic et. al. 2014 DOI: 10.1063
	cdef np.complex128_t *auxbJov
	cdef np.complex128_t *aux3C
	cdef np.complex128_t *Vand
	cdef np.complex128_t *P
	cdef np.complex128_t *Pinv
	cdef np.complex128_t *q

	auxbJov = <np.complex128_t*>malloc(nr*sizeof(np.complex128_t))
	aux3C   = <np.complex128_t*>malloc(nr*nr*sizeof(np.complex128_t))
	aux4C   = <np.complex128_t*>malloc(nr*nr*sizeof(np.complex128_t))
	Vand    = <np.complex128_t*>malloc((nr*(n-1))*sizeof(np.complex128_t))
	P       = <np.complex128_t*>malloc(nr*nr*sizeof(np.complex128_t))
	Pinv    = <np.complex128_t*>malloc(nr*nr*sizeof(np.complex128_t))
	q       = <np.complex128_t*>malloc(nr*sizeof(np.complex128_t))

	cr_start('DMD.amplitudes', 0)
	c_zvandermonde(Vand, auxmuReal, auxmuImag, nr, n-1)
	c_zmatmult(aux3C, w, w, nr, nr, nr, 'C', 'N')
	c_zmatmult(aux4C, Vand, Vand, nr, nr, n-1, 'N', 'C')

	for irow in range(nr):
		for icol in range(nr): #Loop on the columns of the Vandermonde matrix
			P[irow*nr + icol]  = creal(aux3C[irow*nr + icol])*creal(aux4C[irow*nr + icol])
			P[irow*nr + icol] += -creal(aux3C[irow*nr + icol])*cimag(aux4C[irow*nr + icol])*J
			P[irow*nr + icol] += cimag(aux3C[irow*nr + icol])*creal(aux4C[irow*nr + icol])*J
			P[irow*nr + icol] += cimag(aux3C[irow*nr + icol])*cimag(aux4C[irow*nr + icol])
	retval = c_zcholesky(P, nr)
	if not retval == 0: raiseError('Problems computing Cholesky factorization!')

	for iaux in range(nr):
		for irow in range(nr):
			aux1C[irow] = 0 + 0*J
			for icol in range(n-1):#casting Vr to a complex, at the same time, it is multipilied per S and Vand
				aux1C[irow] += Sr[irow]*Vr[irow*(n-1) + icol]*(creal(Vand[iaux*(n-1) + icol])+cimag(Vand[iaux*(n-1) + icol])*J)
			aux2C[irow] = w[irow*nr + iaux]
		c_zmatmult(&q[iaux], aux1C, aux2C, 1, 1, nr, 'N', 'N')

	memcpy(Pinv, P, nr*nr*sizeof(np.complex128_t))
	cdef int ii
	cdef int jj
	for ii in range(nr):
		q[ii] = creal(q[ii]) - cimag(q[ii])*J
		for jj in range(nr - ii):
			P[ii*nr + ii+jj]   = creal(P[(ii+jj)*nr + ii])  - cimag(P[(ii+jj)*nr + ii])*J
			P[(ii+jj)*nr + ii] = creal(Pinv[ii*nr + ii+jj]) - cimag(Pinv[ii*nr + ii+jj])*J

	retval = c_zinverse(Pinv, nr, 'L')
	if not retval == 0: raiseError('Problems computing the Inverse!')

	c_zmatmult(aux1C, Pinv, q, nr, 1, nr, 'N', 'N')

	retval = c_zinverse(P, nr, 'U')
	if not retval == 0: raiseError('Problems computing the Inverse!')

	c_zmatmult(auxbJov, P, aux1C, nr, 1, nr, 'N', 'N')
	cr_stop('DMD.amplitudes',0)

	# Free allocated arrays before reordering
	free(Ur)
	free(Sr)
	free(Vr)
	free(aux1C)
	free(aux2C)
	free(aux3C)
	free(aux4C)
	free(w)
	free(Vand)
	free(q)
	free(P)
	free(Pinv)

	# Order modes and eigenvalues according to its amplitude
	cdef int *auxOrd
	auxOrd = <int*>malloc(nr*sizeof(int))
	cdef np.ndarray[np.double_t,ndim=1] muReal   = np.zeros((nr),dtype=np.double)
	cdef np.ndarray[np.double_t,ndim=1] muImag   = np.zeros((nr),dtype=np.double)
	cdef np.ndarray[np.complex128_t,ndim=2] Phi  = np.zeros((m,nr),order='C',dtype=np.complex128)
	cdef np.ndarray[np.complex128_t,ndim=1] bJov = np.zeros((nr,),dtype=np.complex128)

	cr_start('DMD.qsort', 0)
	c_zsort(auxbJov, auxOrd, nr)
	cr_stop('DMD.qsort', 0)
	cr_start('DMD.sort', 0)
	for ii in range(nr):
		muReal[nr-(auxOrd[ii]+1)] = auxmuReal[ii]
		muImag[nr-(auxOrd[ii]+1)] = auxmuImag[ii]
		bJov[nr-(auxOrd[ii]+1)]   = auxbJov[ii]
		for jj in range(m):
			Phi[jj,nr-(auxOrd[ii]+1)]  = auxPhi[jj*nr + ii]
	cr_stop('DMD.sort', 0)

	# Free the variables that had to be ordered
	free(auxmuReal)
	free(auxmuImag)
	free(auxbJov)
	free(auxPhi)
	free(auxOrd)

	# Ensure that all conjugate modes are in the same order
	cr_start('DMD.conjugate', 0)
	cdef bint p = 0
	cdef double iimag
	for ii in range(nr):
		if p == 1:
			p = 0
			continue
		iimag = muImag[ii]
		if iimag < 0:
			muImag[ii]   =  muImag[ii+1]
			muImag[ii+1] = -muImag[ii]
			bJov[ii]     = creal(bJov[ii])   + cimag(bJov[ii+1])*J
			bJov[ii+1]   = creal(bJov[ii+1]) - cimag(bJov[ii])*J
			for jj in range(m):
				Phi[jj,ii]   = creal(Phi[jj,ii])   + cimag(Phi[jj,ii+1])*J
				Phi[jj,ii+1] = creal(Phi[jj,ii+1]) - cimag(Phi[jj,ii+1])*J
			p = 1
			continue
		if iimag > 0:
			p = 1
			continue
	cr_stop('DMD.conjugate', 0)
	
	# Return
	return muReal, muImag, Phi, bJov

@cr('DMD.run')
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def run(real[:,:] X, real r, int remove_mean=True):
	'''
	Run DMD analysis of a matrix X.

	Inputs:
		- X[ndims*nmesh,n_temp_snapshots]: data matrix
		- remove_mean:                     whether or not to remove the mean flow
		- r:                               maximum truncation residual

	Returns:
		- Phi:      DMD Modes
		- muReal:   Real part of the eigenvalues
		- muImag:   Imaginary part of the eigenvalues
		- b:        Amplitude of the DMD modes
		- Variables needed to reconstruct flow
	'''
	if real is double:
		return _drun(X,r,remove_mean)
	else:
		return _srun(X,r,remove_mean)


## DMD frequency damping
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _sfrequency_damping(float[:] rreal, float[:] iimag, float dt):
	'''
	Computation of the damping ratio and the frequency of each mode
	'''
	cdef int ii, n = rreal.shape[0]
	cdef float mod
	cdef float arg

	cdef np.ndarray[np.float_t,ndim=1] delta = np.zeros((n),dtype=np.float)
	cdef np.ndarray[np.float_t,ndim=1] omega = np.zeros((n),dtype=np.float)

	for ii in range(n):
		mod       = sqrt(rreal[ii]*rreal[ii] + iimag[ii]*iimag[ii])
		delta[ii] = log(mod)/dt
		arg       = atan2(iimag[ii], rreal[ii])
		omega[ii] = arg/dt
	return delta, omega

@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _dfrequency_damping(double[:] rreal, double[:] iimag, double dt):
	'''
	Computation of the damping ratio and the frequency of each mode
	'''
	cdef int ii, n = rreal.shape[0]
	cdef double mod
	cdef double arg

	cdef np.ndarray[np.double_t,ndim=1] delta = np.zeros((n),dtype=np.double)
	cdef np.ndarray[np.double_t,ndim=1] omega = np.zeros((n),dtype=np.double)

	for ii in range(n):
		mod       = sqrt(rreal[ii]*rreal[ii] + iimag[ii]*iimag[ii])
		delta[ii] = log(mod)/dt
		arg       = atan2(iimag[ii], rreal[ii])
		omega[ii] = arg/dt
	return delta, omega

@cr('DMD.frequency_damping')
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def frequency_damping(real[:] rreal, real[:] iimag, real dt):
	'''
	Computation of the damping ratio and the frequency of each mode
	'''
	if real is double:
		return _dfrequency_damping(rreal,iimag,dt)
	else:
		return _sfrequency_damping(rreal,iimag,dt)


## Flow reconstruction
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _sreconstruction_jovanovic(np.complex64_t[:,:] Phi, float[:] muReal, float[:] muImag, float[:] t, np.complex64_t[:] bJov):
	'''
	Computation of the reconstructed flow from the DMD computations
	'''
	cdef int m  = Phi.shape[0], n  = t.shape[0], nr = Phi.shape[1]
	cdef np.complex64_t *Vand
	cdef np.ndarray[np.complex64_t,ndim=2] Zdmd = np.zeros((m,n),order='C',dtype=np.complex64)

	Vand = <np.complex64_t*>malloc(nr*n*sizeof(np.complex64_t))

	c_cvandermonde_time(Vand, &muReal[0], &muImag[0], nr, n, &t[0])
	c_cvecmat(&bJov[0], Vand, nr, n)
	c_cmatmult(&Zdmd[0,0], &Phi[0,0], Vand, m, n, nr, 'N', 'N')
	
	free(Vand)

	return Zdmd.real

@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _dreconstruction_jovanovic(np.complex128_t[:,:] Phi, double[:] muReal, double[:] muImag, double[:] t, np.complex128_t[:] bJov):
	'''
	Computation of the reconstructed flow from the DMD computations
	'''
	cdef int m  = Phi.shape[0], n  = t.shape[0], nr = Phi.shape[1]
	cdef np.complex128_t *Vand
	cdef np.ndarray[np.complex128_t,ndim=2] Zdmd = np.zeros((m,n),order='C',dtype=np.complex128)

	Vand = <np.complex128_t*>malloc(nr*n*sizeof(np.complex128_t))

	c_zvandermonde_time(Vand, &muReal[0], &muImag[0], nr, n, &t[0])
	c_zvecmat(&bJov[0], Vand, nr, n)
	c_zmatmult(&Zdmd[0,0], &Phi[0,0], Vand, m, n, nr, 'N', 'N')
	
	free(Vand)

	return Zdmd.real

@cr('DMD.reconstruction_jovanovic')
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def reconstruction_jovanovic(real_complex[:,:] Phi, real[:] muReal, real[:] muImag, real[:] t, real_complex[:] bJov):
	'''
	Computation of the reconstructed flow from the DMD computations
	'''
	if real_complex is np.complex128_t:
		return _dreconstruction_jovanovic(Phi,muReal,muImag,t,bJov)
	else:
		return _sreconstruction_jovanovic(Phi,muReal,muImag,t,bJov)


#!/usr/bin/env python
#
# PYLOM Testsuite
# Run MLP on the synthetic dataset
#
# Last revision: 23/10/2024
from __future__ import print_function, division

import sys, os,torch
import pyLOM, pyLOM.NN


DATAFILE  = sys.argv[1]
VARIABLES = eval(sys.argv[2])
OUTDIR    = sys.argv[3]


## Set device
device = pyLOM.NN.select_device('cpu')


## Data loading
d = pyLOM.Dataset.load(DATAFILE)
y = d.X(*VARIABLES)
print(d)


## Load pyLOM dataset and set up results output
RESUDIR = os.path.join(OUTDIR,f'MLP_{DATAFILE}')
pyLOM.NN.create_results_folder(RESUDIR,verbose=False)


## Generate torch dataset
input_scaler  = pyLOM.NN.MinMaxScaler()
output_scaler = pyLOM.NN.MinMaxScaler()

dataset = pyLOM.NN.Dataset(
    variables_out       = (y,), 
    variables_in        = d.xyz,
    parameters          = [d.get_variable('Re'), d.get_variable('AoA')],
    inputs_scaler       = input_scaler,
    outputs_scaler      = output_scaler,
    snapshots_by_column = True
)

td_train, td_test = dataset.get_splits([0.8, 0.2])


## Generate model
training_params = {
    "epochs": 50,
    "lr": 0.00015,
    "lr_gamma": 0.98,
    "lr_scheduler_step": 15,
    "batch_size": 512,
    "loss_fn": torch.nn.MSELoss(),
    "optimizer_class": torch.optim.Adam,
    "print_rate_epoch": 10,
}

sample_input, sample_output = td_train[0]
model = pyLOM.NN.MLP(
    input_size=sample_input.shape[0],
    output_size=sample_output.shape[0],
    hidden_size=32,
    n_layers=2,
    p_dropouts=0.1,
)

pipeline = pyLOM.NN.Pipeline(
    train_dataset=td_train,
    test_dataset=td_test,
    model=model,
    training_params=training_params,
)

training_logs = pipeline.run()


## check saving and loading the model
pipeline.model.save(os.path.join(RESUDIR,"model.pth"))
model = pyLOM.NN.MLP.load(RESUDIR + "/model.pth")
preds = model.predict(td_test, batch_size=250)

scaled_preds = output_scaler.inverse_transform([preds])[0]
scaled_y     = output_scaler.inverse_transform([td_test[:][1].numpy()])[0]


## Testsuite output
pyLOM.pprint(0,'TSUITE y            =',y.min(),y.max(),y.mean())
#pyLOM.pprint(0,'TSUITE scaled_y     =',scaled_y.min(),scaled_y.max(),scaled_y.mean())
#pyLOM.pprint(0,'TSUITE scaled_preds =',scaled_preds.min(),scaled_preds.max(),scaled_preds.mean())

pyLOM.cr_info()
pyLOM.pprint(0,'End of output')


#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# Encoder-Decoder architecture for NN Module
#
# Last rev: 09/10/2024

import torch.nn as nn


class Encoder2D(nn.Module):
    r"""
    Encoder2D class for the 2D Convolutional Autoencoder.

    Args:
        nlayers (int): Number of layers in the encoder.
        latent_dim (int): Latent dimension of the encoder.
        nh (int): Height of the input mesh/image.
        nw (int): Width of the input mesh/image.
        input_channels (int): Number of input channels.
        filter_channels (int): Number of filter channels.
        kernel_size (int): Kernel size for the convolutional layers.
        padding (int): Padding for the convolutional layers.
        activation_funcs (list): List of activation functions.
        nlinear (int): Number of neurons in the linear layer.
        batch_norm (bool): Whether to use batch normalization. Default is ``False``.
        stride (int): Stride for the convolutional layers. Default is ``2``.
        dropout (float): Dropout probability. Default is ``0``.
        vae (bool): Wheather the encoder is going to be used on a VAE or not. Default is ``False``.
    """
    def __init__(
        self,
        nlayers: int,
        latent_dim: int,
        nh: int,
        nw: int,
        input_channels: int,
        filter_channels: int,
        kernel_size: int,
        padding: int,
        activation_funcs: list,
        nlinear: int,
        batch_norm: bool = False,
        stride: int = 2,
        dropout: float = 0,
        vae: bool = False,
    ):
        super(Encoder2D, self).__init__()

        self.nlayers    = nlayers
        self.filt_chan  = filter_channels
        self.in_chan    = input_channels
        self.lat_dim    = latent_dim
        self.nh         = nh
        self.nw         = nw
        self.isvae      = vae
        self.funcs      = activation_funcs
        self.nlinear    = nlinear
        self.batch_norm = batch_norm
        self.dropout    = nn.Dropout(p=dropout)

        # Create a list to hold the convolutional layers
        self.conv_layers = nn.ModuleList()
        self.norm_layers = nn.ModuleList()
        in_channels = self.in_chan # Initial input channels
        for ilayer in range(self.nlayers):
            out_channels = self.filt_chan * (1 << ilayer)  # Compute output channels
            conv_layer = nn.Conv2d(in_channels, out_channels, kernel_size, stride, padding)
            self.conv_layers.append(conv_layer)
            if self.batch_norm:
                self.norm_layers.append(nn.BatchNorm2d(out_channels))
            in_channels = out_channels  # Update in_channels for the next layer
       
        self.flat     = nn.Flatten()
        fc_input_size = out_channels * (self.nh // (1 << self.nlayers)) * (self.nw // (1 << self.nlayers))
        self.fc1      = nn.Linear(fc_input_size, self.nlinear)
        if self.isvae:
            self.mu     = nn.Linear(self.nlinear, self.lat_dim)
            self.logvar = nn.Linear(self.nlinear, self.lat_dim)
        else:
            self.z = nn.Linear(self.nlinear, self.lat_dim)

        self._reset_parameters()
    
    def _reset_parameters(self):
        for layer in self.modules():
            if isinstance(layer, nn.Conv2d):
                nn.init.xavier_uniform_(layer.weight)
            elif isinstance(layer, nn.Linear):
                nn.init.xavier_uniform_(layer.weight)

    def forward(self, x):  
        out = x
        for ilayer, conv_layer in enumerate(self.conv_layers):
            out = conv_layer(out)
            if self.batch_norm:
                out = self.norm_layers[ilayer](out)
            out = self.funcs[ilayer](out)
        out = self.funcs[ilayer+1](self.flat(out))
        out = self.funcs[ilayer+2](self.fc1(out))
        if self.isvae:
            return self.mu(out), self.logvar(out)
        else:
            return self.z(out)


class Decoder2D(nn.Module):
    r"""
    Decoder2D class for the 2D Convolutional Autoencoder.

    Args:
        nlayers (int): Number of layers in the encoder.
        latent_dim (int): Latent dimension of the encoder.
        nh (int): Height of the input mesh/image.
        nw (int): Width of the input mesh/image.
        input_channels (int): Number of input channels.
        filter_channels (int): Number of filter channels.
        kernel_size (int): Kernel size for the convolutional layers.
        padding (int): Padding for the convolutional layers.
        activation_funcs (list): List of activation functions.
        nlinear (int): Number of neurons in the linear layer.
        batch_norm (bool): Whether to use batch normalization. Default is ``False``.
        stride (int): Stride for the convolutional layers. Default is ``2``.
        dropout (float): Dropout probability. Default is ``0``.
    """
    def __init__(
        self,
        nlayers: int,
        latent_dim: int,
        nh: int,
        nw: int,
        input_channels: int,
        filter_channels: int,
        kernel_size: int,
        padding: int,
        activation_funcs: list,
        nlinear: int,
        batch_norm: bool = False,
        stride: int = 2,
        dropout: float = 0,
    ):
        super(Decoder2D, self).__init__()       
        
        self.nlayers    = nlayers
        self.filt_chan  = filter_channels
        self.in_chan    = input_channels
        self.lat_dim    = latent_dim
        self.nh         = nh
        self.nw         = nw
        self.funcs      = activation_funcs
        self.nlinear    = nlinear
        self.batch_norm = batch_norm
        self.dropout    = nn.Dropout(p=dropout)
        self.fc1 = nn.Linear(in_features=self.lat_dim, out_features=self.nlinear)
        fc_output_size = int((self.filt_chan * (1 << (self.nlayers-1)) * self.nh // (1 << self.nlayers) * self.nw // (1 << self.nlayers)))
        self.fc2 = nn.Linear(in_features=self.nlinear, out_features=fc_output_size)

        # Create a list to hold the transposed convolutional layers
        self.deconv_layers = nn.ModuleList()
        self.norm_layers   = nn.ModuleList()
        in_channels = self.filt_chan * (1 << self.nlayers-1)
        for i in range(self.nlayers-1, 0, -1):
            out_channels = self.filt_chan * (1 << (i - 1))  # Compute output channels
            deconv_layer = nn.ConvTranspose2d(in_channels, out_channels, kernel_size, stride, padding)
            self.deconv_layers.append(deconv_layer)
            if self.batch_norm:
                self.norm_layers.append(nn.BatchNorm2d(in_channels))
            in_channels = out_channels  # Update in_channels for the next layer
        out_channels = self.in_chan
        deconv_layer = nn.ConvTranspose2d(in_channels, out_channels, kernel_size, stride, padding)
        self.deconv_layers.append(deconv_layer)
        if self.batch_norm:
            self.norm_layers.append(nn.BatchNorm2d(in_channels))

        self._reset_parameters()

    def _reset_parameters(self):
        for layer in self.modules():
            if isinstance(layer, nn.ConvTranspose2d):
                nn.init.xavier_uniform_(layer.weight)
            elif isinstance(layer, nn.Linear):
                nn.init.xavier_uniform_(layer.weight)

    def forward(self, x):
        out = self.funcs[self.nlayers+1](self.fc1(x))
        out = self.funcs[self.nlayers](self.fc2(out))
        out = out.view(out.size(0), self.filt_chan * (1 << (self.nlayers-1)), int(self.nh // (1 << self.nlayers)), int(self.nw // (1 << self.nlayers)))
        for ilayer, (deconv_layer) in enumerate(self.deconv_layers[:-1]):
            if self.batch_norm:
                out = self.norm_layers[ilayer](out)
            out = self.funcs[self.nlayers-ilayer-1](deconv_layer(out))
        return self.deconv_layers[-1](out)


class Encoder3D(nn.Module):
    r"""
    Encoder3D class for the 3D Convolutional Encoder.
    
    Args:
        nlayers (int): Number of layers in the encoder.
        latent_dim (int): Latent dimension of the encoder.
        nx (int): Height of the input mesh/image.
        ny (int): Width of the input mesh/image.
        nz (int): Depth of the input mesh/image.
        input_channels (int): Number of input channels.
        filter_channels (int): Number of filter channels.
        kernel_size (int): Kernel size for the convolutional layers.
        padding (int): Padding for the convolutional layers.
        activation_funcs (list): List of activation functions.
        nlinear (int): Number of neurons in the linear layer.
        batch_norm (bool): Whether to use batch normalization. Default is ``False``.
        stride (int): Stride for the convolutional layers. Default is ``2``.
        dropout (float): Dropout probability. Default is ``0``.
        vae (bool): Wheather the encoder is going to be used on a VAE or not. Default is ``False``.
    """
    def __init__(
        self,
        nlayers: int,
        latent_dim: int,
        nx: int,
        ny: int,
        nz: int,
        input_channels: int,
        filter_channels: int,
        kernel_size: int,
        padding: int,
        activation_funcs: list,
        nlinear: int,
        batch_norm: bool = False,
        stride: int = 2,
        dropout: float = 0,
        vae: bool = False,
    ):
        super(Encoder3D,self).__init__()

        self.nlayers = nlayers
        self.filt_chan = filter_channels
        self.in_chan = input_channels
        self._lat_dim = latent_dim
        self._nx = nx
        self._ny = ny
        self._nz = nz
        self._isvae = vae
        self.funcs = activation_funcs
        self.nlinear = nlinear
        self.batch_norm = batch_norm
        self.dropout = nn.Dropout(p=dropout)

        # List to hold the Conv Layers

        self.conv_layers = nn.ModuleList()
        self.norm_layers = nn.ModuleList()
        in_channels = self.in_chan #Initial input channels
        for ilayer in range(self.nlayers):
            out_channels = self.filt_chan * (1 << ilayer) #Compute output channels using the shift operator
            conv_layer = nn.Conv3d(in_channels, out_channels, kernel_size, stride, padding)
            self.conv_layers.append(conv_layer)
            if self.batch_norm:
                self.norm_layers.append(nn.BatchNorm3d(out_channels))
            in_channels = out_channels # Update n of channels for the next layer

        self.flat = nn.Flatten()
        fc_input_size = out_channels * (self._nx // (1<<self.nlayers)) * (self._ny // (1<<self.nlayers)) * (self._nz // (1<<self.nlayers)) #Compute FullyConnected layer input size
        self.fc1 = nn.Linear(fc_input_size, self.nlinear)
        if self._isvae:
            self.mu = nn.Linear(self.nlinear, self._lat_dim)
            self.logvar = nn.Linear(self.nlinear, self._lat_dim)
        else:
            self.z = nn.Linear(self.nlinear, self._lat_dim)

        self._reset_parameters()
    
    def _reset_parameters(self):
        for layer in self.modules():
            if isinstance(layer, nn.Conv3d):
                nn.init.xavier_uniform_(layer.weight)
            elif isinstance(layer, nn.Linear):
                nn.init.xavier_uniform_(layer.weight)
    
    def forward(self, x):
        out = x
        for ilayer, conv_layer in enumerate(self.conv_layers):
            out = conv_layer(out)
            if self.batch_norm:
                out=self.norm_layers[ilayer](out)
            out = self.funcs[ilayer](out)
        out = self.funcs[ilayer+1](self.flat(out))
        out = self.funcs[ilayer+2](self.fc1(out))
        if self._isvae:
            return self.mu(out), self.logvar(out)
        else:
            return self.z(out)


class Decoder3D(nn.Module):
    r"""
    Dencoder3D class for the 3D Convolutional Autoencoder.

    Args:
        nlayers (int): Number of layers in the encoder.
        latent_dim (int): Latent dimension of the encoder.
        nx (int): Height of the input mesh/image.
        ny (int): Width of the input mesh/image.
        nz (int): Depth of the input mesh/image.
        input_channels (int): Number of input channels.
        filter_channels (int): Number of filter channels.
        kernel_size (int): Kernel size for the convolutional layers.
        padding (int): Padding for the convolutional layers.
        activation_funcs (list): List of activation functions.
        nlinear (int): Number of neurons in the linear layer.
        batch_norm (bool): Whether to use batch normalization. Default is ``False``.
        stride (int): Stride for the convolutional layers. Default is ``2``.
        dropout (float): Dropout probability. Default is ``0``.
    """
    def __init__(
        self,
        nlayers: int,
        latent_dim: int,
        nx: int,
        ny: int,
        nz: int,
        input_channels: int,
        filter_channels: int,
        kernel_size: int,
        padding: int,
        activation_funcs: list,
        nlinear: int,
        batch_norm: bool = False,
        stride: int = 2,
        dropout: float = 0,
    ):
        
        super(Decoder3D, self).__init__()       
        
        self.nlayers = nlayers
        self.filt_chan = filter_channels
        self.in_chan = input_channels
        self.lat_dim = latent_dim
        self.nx = nx
        self.ny = ny
        self.nz = nz
        self.funcs = activation_funcs
        self.nlinear = nlinear
        self.batch_norm = batch_norm
        self.dropout = nn.Dropout(p=dropout)
        self.fc1 = nn.Linear(in_features=self.lat_dim, out_features=self.nlinear)
        fc_output_size = int((self.filt_chan * (1 << (self.nlayers-1)) * self.nx // (1 << self.nlayers) * self.ny // (1 << self.nlayers) * self.nz // (1 << self.nlayers)))
        self.fc2 = nn.Linear(in_features=self.nlinear, out_features=fc_output_size)

        # List to hold the transposed convolutional layers
        self.deconv_layers = nn.ModuleList()
        self.norm_layers   = nn.ModuleList()
        in_channels = self.filt_chan * (1 << self.nlayers-1)
        for i in range(self.nlayers-1, 0, -1):
            out_channels = self.filt_chan * (1 << (i - 1))  # Compute output channels
            deconv_layer = nn.ConvTranspose3d(in_channels, out_channels, kernel_size, stride, padding)
            self.deconv_layers.append(deconv_layer)
            if self.batch_norm:
                self.norm_layers.append(nn.BatchNorm3d(in_channels))
            in_channels = out_channels  # Update in_channels for the next layer
        out_channels = self.in_chan
        deconv_layer = nn.ConvTranspose3d(in_channels, out_channels, kernel_size, stride, padding)
        self.deconv_layers.append(deconv_layer)
        if self.batch_norm:
            self.norm_layers.append(nn.BatchNorm3d(in_channels))

        self._reset_parameters()

    def _reset_parameters(self):
        for layer in self.modules():
            if isinstance(layer, nn.ConvTranspose3d):
                nn.init.xavier_uniform_(layer.weight)
            elif isinstance(layer, nn.Linear):
                nn.init.xavier_uniform_(layer.weight)

    def forward(self, x):
        out = self.funcs[self.nlayers+1](self.fc1(x))
        out = self.funcs[self.nlayers](self.fc2(out))
        out = out.view(out.size(0), self.filt_chan * (1 << (self.nlayers-1)), int(self.nx // (1 << self.nlayers)), int(self.ny // (1 << self.nlayers)), int(self.nz // (1 << self.nlayers)))
        for ilayer, (deconv_layer) in enumerate(self.deconv_layers[:-1]):
            if self.batch_norm:
                out = self.norm_layers[ilayer](out)
            out = self.funcs[self.nlayers-ilayer-1](deconv_layer(out))
        return self.deconv_layers[-1](out)



#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# NN Module
#
# Last rev: 09/10/2024

# Supress prints from tensorflow
import os, torch, torch.nn as nn
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'  
DEVICE = "cuda" if torch.cuda.is_available() else "cpu"

from ..utils.plots  import plotSnapshot

from .pipeline      import Pipeline
from .utils         import Dataset, MinMaxScaler, select_device, betaLinearScheduler, create_results_folder, set_seed

from .optimizer     import OptunaOptimizer

from .stats         import RegressionEvaluator
from .callbacks     import EarlyStopper

from .architectures.mlp               import MLP
from .architectures.kan               import KAN, ChebyshevLayer, JacobiLayer
from .architectures.autoencoders      import Autoencoder, VariationalAutoencoder
from .architectures.encoders_decoders import Encoder2D, Decoder2D, Encoder3D, Decoder3D
from .architectures.pinn              import PINN, BurgersPINN, Euler2DPINN, NavierStokesIncompressible2DPINN, BoundaryCondition



# Wrapper of the activation functions
def tanh():      return nn.Tanh()
def relu():      return nn.ReLU()
def elu():       return nn.ELU()
def sigmoid():   return nn.Sigmoid()
def leakyRelu(): return nn.LeakyReLU()
def silu():      return nn.SiLU()

del os, torch



#!/usr/bin/env python
#
# pyLOM, dataset.
#
# Dataset class, reader and reduction routines.
#
# Last rev: 30/07/2021
from __future__ import print_function, division

import os, numpy as np

from .      import inp_out as io
from .utils import cr_nvtx as cr, mem, raiseError, gpu_to_cpu, cpu_to_gpu


class Dataset(object):
	'''
	The Dataset class wraps the position of the nodes and the time instants
	with the number of variables and relates them so that the operations 
	in parallel are easier.
	'''
	def __init__(self, xyz=None, ptable=None, vars=None, order=None, point=True, **kwargs):
		'''
		Class constructor

		Inputs:
			> xyz:    coordinates of the points.
			> ptable: partition table used.
			> vars:   dictionary containing the variable name and values as
					  as a python dictionary.
			> order:  ordering of the points (automatically created if none)
			> point:  True if point data, False if cell data.
			> kwags:  dictionary containing the field name and values as a
					  python dictionary.
		'''
		self._xyz      = xyz
		self._vardict  = vars
		self._fieldict = kwargs
		self._ptable   = ptable
		self._order    = np.arange(xyz.shape[0]) if order is None else order
		self._point    = point

	def __len__(self):
		return self._xyz.shape[0]

	def __str__(self):
		'''
		String representation
		'''
		s  = 'Dataset of %d variables:\n' % len(self.varnames)
		for key in self.varnames:
			var    = self.vars[key]['value']
			nanstr = ' (has NaNs) ' if np.any(np.isnan(var)) else ' '
			s     += '  > ' + key + nanstr + ' - max = ' + str(np.nanmax(var)) + ', min = ' + str(np.nanmin(var)) + '\n'
		s += 'and %d fields with %d points:\n' % (len(self.fieldnames),len(self))
		for key in self.fieldnames:
			field  = self.fields[key]['value']
			nanstr = ' (has NaNs) ' if np.any(np.isnan(field)) else ' '
			s     += '  > ' +  key + nanstr + '- max = ' + str(np.nanmax(field)) \
										    + ', min = ' + str(np.nanmin(field)) \
										    + ', avg = ' + str(np.nanmean(field)) \
										    + '\n'
		return s
		
	# Set and get functions
	def __getitem__(self,key):
		'''
		Dataset[key]

		Recover the value of a field given its key
		'''
		return self._fieldict[key]['value']

	def __setitem__(self,key,value):
		'''
		Dataset[key] = value

		Set the field of a variable given its key
		'''
		self._fieldict[key]['value'] = value

	# Functions
	def rename(self,new,old):
		'''
		Rename a variable inside a field.
		'''
		self.fields[new] = self.fields.pop(old)
		return self

	def delete(self,varname):
		'''
		Delete a variable inside a field.
		'''
		return self.fields.pop(varname)

	def get_variable(self,key):
		'''
		Recover the value of a variable given its key
		'''
		return self._vardict[key]['value']

	def set_variable(self,key,value):
		'''
		Recover the value of a variable given its key
		'''
		self._vardict[key]['value'] = value

	def get_dim(self,var,idim):
		'''
		Recover the value of a variable for a given dimension
		'''
		ndim = self._fieldict[var]['ndim']
		if idim >= ndim: raiseError(f'Requested dimension {idim} for {var} greater than its number of dimensions {ndim}!')
		print(len(self))
		return  np.ascontiguousarray(self._fieldict[var]['value'][idim:ndim*len(self):ndim])

	def info(self,var):
		'''
		Returns the information for a certain variable
		'''
		return {'point':self._point,'ndim':self._fieldict[var]['ndim']}
	
	def to_gpu(self,fields=None):
		'''
		Send field data to the GPU
		'''
		fields = fields if not fields is None else self.fieldnames
		for key in fields:
			self._fieldict[key]['value'] = cpu_to_gpu(self._fieldict[key]['value'])
		return self

	def to_cpu(self,fields=None):
		'''
		Send field data to the CPU
		'''
		fields = fields if not fields is None else self.fieldnames
		for key in fields:
			self._fieldict[key]['value'] = gpu_to_cpu(self._fieldict[key]['value'])
		return self

	def add_field(self,varname,ndim,var):
		'''
		Add a field to the dataset
		'''
		self._fieldict[varname] = {
			'ndim'  : ndim,
			'value' : var, 
		}

	def add_variable(self,varname,idim,var):
		'''
		Add a variable to the dataset
		'''
		self._vardict[varname] = {
			'idim'  : idim,
			'value' : var, 
		}

	def append_variable(self,varname,var,**fieldict):
		'''
		Appends new timesteps to the dataset
		'''
		# Add to variable vector
		self.vars[varname]['value'] = np.concatenate((self.vars[varname]['value'],var))
		# Sort ascendingly and retrieve sorting index
		idx = np.argsort(self.vars[varname]['value'])
		self.vars[varname]['value'] = self.vars[varname]['value'][idx]
		idim = self.vars[varname]['idim']
		# Now concatenate and sort per variable
		for v in fieldict:
			aux = np.concatenate((self[v][:,:,idim],fieldict[v]),axis=1)[:,idx]
			self[v][:,:,idim] = aux

	@cr('Dataset.reshape')
	def reshape(self,field,info):
		'''
		Reshape a field for a single variable
		according to the info
		'''
		# Obtain number of points from the mesh
		npoints = len(self)
		# Only reshape the variable if ndim > 1
		return np.ascontiguousarray(field.reshape((npoints,info['ndim']),order='C') if info['ndim'] > 1 else field)

	@cr('Dataset.X')
	def X(self,*args):
		'''
		Return the X matrix for the selected fields
		'''
		# Select all variables if none is provided
		fieldnames = self.fieldnames if len(args) == 0 else args
		# Compute the number of fields
		npoints = len(self)
		nfields = 0
		for f in fieldnames:
			nfields += self.fields[f]['ndim']
		dims = [nfields*npoints]
		# Variable order could be random, thus create a list of variable
		# names and their idim to order
		varls = np.array(list(self.varnames))
		ivars = np.array([self.vars[v]['idim'] for v in varls])
		idx   = np.argsort(ivars)
		# Order the variables
		varls = varls[idx]
		ivars = ivars[idx]
		# Loop the number of variables according to their idim
		# As minimum we will have 1 variable, thus idim=0. If 
		# we have idim > 0, this surely indicates a multi-dimensional
		# field
		varc = 0
		for v in varls:
			ivar = self.vars[v]['idim']
			lvar = len(self.vars[v]['value'])
			if ivar == varc:
				dims += [lvar]
				varc += 1
		# Create output array
		X = np.zeros(dims,np.double)
		# Populate output matrix
		ifield = 0
		for field in fieldnames:
			v = self.fields[field]
			for idim in range(v['ndim']):
				X[ifield:nfields*npoints:nfields] = v['value'][idim:v['ndim']*npoints:v['ndim']]
				ifield += 1
		return X

	@cr('Dataset.save')
	def save(self,fname,**kwargs):
		'''
		Store the field in various formats.
		'''
		# Guess format from extension
		fmt = os.path.splitext(fname)[1][1:] # skip the .
		# Pickle format
		if fmt.lower() == 'pkl': 
			io.pkl_save(fname,self)
		# H5 format
		if fmt.lower() == 'h5':
			# Set default parameters
			if not 'mode' in kwargs.keys():        kwargs['mode']        = 'w' if not os.path.exists(fname) else 'a'
			if not 'mpio' in kwargs.keys():        kwargs['mpio']        = True
			if not 'nopartition' in kwargs.keys(): kwargs['nopartition'] = False
			# Append or save
			if not kwargs.pop('append',False):
				io.h5_save_dset(fname,self.xyz,self.vars,self.fields,self.ordering,self.point,self.partition_table,**kwargs)
			else:
				io.h5_append_dset(fname,self.xyz,self.vars,self.fields,self.ordering,self.point,self.partition_table,**kwargs)

	@classmethod
	@cr('Dataset.load')
	def load(cls,fname,**kwargs):
		'''
		Load a field from various formats
		'''
		# Guess format from extension
		fmt = os.path.splitext(fname)[1][1:] # skip the .
		# Pickle format
		if fmt.lower() == 'pkl': 
			return io.pkl_load(fname)
		# H5 format
		if fmt.lower() == 'h5':
			if not 'mpio' in kwargs.keys(): kwargs['mpio'] = True
			xyz, order, point, ptable, varDict, fieldDict = io.h5_load_dset(fname,**kwargs)
			return cls(xyz,ptable,varDict,order, point, **fieldDict)
		raiseError('Cannot load file <%s>!'%fname)

	# Properties
	@property
	def xyz(self):
		return self._xyz
	def x(self):
		return self._xyz[:,0]
	@property
	def y(self):
		return self._xyz[:,1]
	@property
	def z(self):
		return self._xyz[:,2]

	@property
	def ordering(self):
		return self._order
	@property
	def point(self):
		return self._point
	@property
	def partition_table(self):
		return self._ptable

	@property
	def vars(self):
		return self._vardict
	@property
	def varnames(self):
		return list(self._vardict.keys())

	@property
	def fields(self):
		return self._fieldict
	@property
	def fieldnames(self):
		return list(self._fieldict.keys())


#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# Utils Module
#
# Last rev: 14/02/2025

from .errors import raiseError, raiseWarning, round
from .cr     import cr, cr_nvtx, cr_start, cr_stop, cr_info
from .nvtxp  import nvtxp
from .mem    import mem, mem_start, mem_stop, mem_info
from .parall import worksplit, writesplit, is_rank_or_serial, pprint
from .mpi    import MPI_COMM, MPI_RANK, MPI_SIZE, mpi_barrier, mpi_send, mpi_recv, mpi_sendrecv, mpi_scatter, mpi_gather, mpi_reduce, mpi_bcast
from .gpu    import gpu_device, gpu_to_cpu, cpu_to_gpu, ascontiguousarray

del errors, parall



#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# Utils - Chrono modules for performance.
#
# Last rev: 09/07/2021
from __future__ import print_function, division

import numpy as np, copy, functools

from .mpi    import MPI, MPI_RANK, MPI_SIZE, mpi_reduce
from .errors import raiseError, raiseWarning


CHANNEL_DICT = {}


class channel(object):
	'''
	This is a channel for the cr counter
	'''
	def __init__(self, name, tmax, tmin, tsum, nop, tini):
		self._name = name # Name of the channel
		self._tmax = tmax # Maximum time of the channel
		self._tmin = tmin # Minimum time of the channel
		self._tsum = tsum # Total time of the channel
		self._nop  = nop  # Number of operations
		self._tini = tini # Initial instant (if == 0 channel is not being take into account)

	def __str__(self):
		return 'name %-30s n %9d tmin %e tmax %e tavg %e tsum %e' % (self.name,self.nop,self.tmin,self.tmax,self.tavg,self.tsum)

	def __add__(self, other):
		new = copy.deepcopy(self)
		new._tmax  = max(new._tmax,other._tmax)
		new._tmin  = min(new._tmin,other._tmin)
		new._tsum += other._tsum
		new._nop  += other._nop 
		return new

	def __iadd__(self, other):
		self._tmax  = max(self._tmax,other._tmax)
		self._tmin  = min(self._tmin,other._tmin)
		self._tsum += other._tsum
		self._nop  += other._nop 
		return self

	def reset(self):
		'''
		Reset the channel
		'''
		self._tmax = 0.0
		self._tmin = 0.0
		self._tsum = 0.0
		self._nop  = 0.0
		self._tini = 0.0

	def restart(self):
		self._tini = 0.0

	def start(self,tini):
		self._tini = tini

	def increase_nop(self):
		self._nop += 1

	def increase_time(self,time):
		self._tsum += time

	def set_max(self,time):
		if time > self._tmax or self._nop == 1: self._tmax = time

	def set_min(self,time):
		if time < self._tmin or self._nop == 1: self._tmin = time

	def elapsed(self,time):
		return time - self._tini

	def is_running(self):
		return not self._tini == 0

	@classmethod
	def new(cls,name):
		'''
		Create a new channel
		'''
		return cls(name,0,0,0,0,0)

	@property
	def name(self):
		return self._name
	@property
	def nop(self):
		return self._nop
	@property
	def tmin(self):
		return self._tmin
	@property
	def tmax(self):
		return self._tmax
	@property
	def tavg(self):
		return self._tsum/(1.* self._nop)
	@property
	def tsum(self):
		return self._tsum


def _newch(ch_name):
	'''
	Add a new channel to the list
	'''
	CHANNEL_DICT[ch_name] = channel.new(ch_name)
	return CHANNEL_DICT[ch_name]

def _findch(ch_name):
	'''
	Look for the channel
	'''
	return CHANNEL_DICT[ch_name] if ch_name in CHANNEL_DICT.keys() else None

def _addsuff(ch_name,suff=-1):
	return ch_name if suff <= 0 else '%s%02d' % (ch_name,suff)

def _findch_crash(ch_name):
	'''
	Look for the channel and crash if it does not exist
	'''
	if not ch_name in CHANNEL_DICT.keys():
		raiseError('Channel %s does not exist!' % ch_name)
	return CHANNEL_DICT[ch_name]

def _findch_create(ch_name):
	'''
	Find the channel and if not found create it
	'''
	return CHANNEL_DICT[ch_name] if ch_name in CHANNEL_DICT.keys() else _newch(ch_name)

def _gettime():
	'''
	Returns the number of second since an arbitrary instant but fixed.
	Returned value will always be > 0.
	'''
	return MPI.Wtime()

def _reduce_cr(cr1,cr2,dtype):
	for key in cr2.keys():
		if key in cr1.keys():
			# Key exists in cr1, then simply accumulate
			cr1[key] += cr2[key]
		else:
			# Key does not exist in cr1, create it new
			cr1[key] = cr2[key]
	return cr1
cr_reduce = MPI.Op.Create(_reduce_cr, commute=True)

def _info_serial():
	tsum_array = np.array([CHANNEL_DICT[key].tsum for key in CHANNEL_DICT.keys()])
	name_array = np.array([CHANNEL_DICT[key].name for key in CHANNEL_DICT.keys()])

	ind = np.argsort(tsum_array) # sorted indices

	print('\ncr_info:',flush=True)
	for ii in ind[::-1]:
		print(CHANNEL_DICT[name_array[ii]],flush=True)
	print('',flush=True)

def _info_parallel():
	CHANNEL_DICT_G = mpi_reduce(CHANNEL_DICT,root=0,op=cr_reduce,all=False)

	if MPI_RANK == 0:
		tsum_array = np.array([CHANNEL_DICT_G[key].tsum for key in CHANNEL_DICT_G.keys()])
		name_array = np.array([CHANNEL_DICT_G[key].name for key in CHANNEL_DICT_G.keys()])	
	
		ind = np.argsort(tsum_array) # sorted indices

		print('\ncr_info (mpi size: %d):' % (MPI_SIZE),flush=True)
		for ii in ind[::-1]:
			print(CHANNEL_DICT_G[name_array[ii]],flush=True)
		print('',flush=True)


def cr_reset():
	'''
	Delete all channels and start again
	'''
	CHANNEL_DICT = {}

def cr_info(rank=-1):
	'''
	Print information - order by major sum
	'''
	if rank >= 0 and rank == MPI_RANK:
		_info_serial()
	else:
		_info_parallel()

def cr_start(ch_name,suff):
	'''
	Start the chrono of a channel
	'''
	name_tmp = _addsuff(ch_name,suff)
	channel  = _findch_create(name_tmp)
	if channel.is_running():
		raiseError('Channel %s was already set!'%channel.name)
	channel.start( _gettime() )

def cr_stop(ch_name,suff):
	'''
	Stop the chrono of a channel
	'''
	end      = _gettime()
	name_tmp = _addsuff(ch_name,suff)
	channel  = _findch_crash(name_tmp)
	time     = channel.elapsed(end)

	channel.increase_nop()
	channel.set_max(time)
	channel.set_min(time)
	channel.increase_time(time)

	channel.restart()

def cr_time(ch_name,suff):
	'''
	Get the time of a channel that is running; channel keeps running
	'''
	end = _gettime()
	name_tmp = _addsuff(ch_name,suff)
	channel  = _findch_crash(name_tmp)
	return channel.elapsed(end)

def cr(ch_name,suff=0):
	'''
	CR decorator
	'''
	def decorator(func):
		@functools.wraps(func)
		def wrapper(*args,**kwargs):
			cr_start(ch_name,suff)
			out = func(*args,**kwargs)
			cr_stop(ch_name,suff)
			return out
		return wrapper
	return decorator

try:
	import nvtx

	def cr_nvtx(ch_name,suff=0,color="green"):
		'''
		CR NVTX decorator
		'''
		def decorator(func):
			@functools.wraps(func)
			def wrapper(*args,**kwargs):
				cr_start(ch_name,suff)
				with nvtx.annotate(message=ch_name,color=color):
					out = func(*args,**kwargs)
				cr_stop(ch_name,suff)
				return out
			return wrapper
		return decorator

except:
	raiseWarning('Import - NVTX not present!',False)

	def cr_nvtx(ch_name,suff=0,color="green"):
		'''
		CR NVTX decorator
		'''
		def decorator(func):
			@functools.wraps(func)
			def wrapper(*args,**kwargs):
				cr_start(ch_name,suff)
				out = func(*args,**kwargs)
				cr_stop(ch_name,suff)
				return out
			return wrapper
		return decorator



#!/usr/bin/env python
#
# Example of GPR.
#
# Last revision: 19/02/2025
from __future__ import print_function, division

import numpy as np
import matplotlib.pyplot as plt
import pyLOM, pyLOM.GPR

# Single Fidelity GPR

# TESTDATA
DATAFILE = './DATA/AIRFOIL.h5'
d = pyLOM.Dataset.load(DATAFILE)

# Data selection
selected_snapshot = 5
pressure_snapshot = d["cp"][:, selected_snapshot]
coordinates = d.xyz

np.random.seed(42)
masked_pressure = pressure_snapshot*(np.random.rand(len(pressure_snapshot)) < 0.65)
nonzero_indices = np.nonzero(masked_pressure)[0]
filtered_coordinates = coordinates[nonzero_indices]

x_train = filtered_coordinates
y_train = masked_pressure[nonzero_indices]
x_pred = coordinates

# GPR
gpr = pyLOM.GPR.SF_GPR(input_dim=2)
pyLOM.pprint(0,gpr.kernel.available_kernels)  # Dict of available kernels
pyLOM.pprint(0,
    gpr.kernel.get_kernel_parameters("RBF")
)  # Dict of available parameters for the selecter kernel
kernel = gpr.kernel.Matern32(variance=0.1, lengthscale=1.0, ARD=True)
gpr.fit(x_train, y_train, kernel)
gpr.display_model()
result = gpr.predict(x_pred)


# PLOT
plt.figure(figsize=(8,6),dpi=100)
plt.plot(x_pred[:, 0], result["mean"], color="navy", label="GPR Mean", zorder=3)
plt.fill_between(
    x_pred[:, 0],
    (result["mean"] - 1.96 * result["std"]).flatten(),
    (result["mean"] + 1.96 * result["std"]).flatten(),
    facecolor="cornflowerblue",
    alpha=0.3,
    zorder=2,
    label="GPR 95% CI",
)
plt.scatter(x_train[:, 0], y_train, color="crimson", label="Train data", zorder=4)
plt.plot(
    coordinates[:, 0], pressure_snapshot, color="darkorange", label="True", zorder=5
)
plt.gca().invert_yaxis()
plt.xlabel("x/c")
plt.ylabel("Cp")
plt.legend()
plt.show()

# Multi Fidelity GPR

# Forrester Function
def forrester_high(x):
    """High-fidelity function"""
    return (6 * x - 2) ** 2 * np.sin(12 * x - 4)


def forrester_low(x):
    """Low-fidelity function"""
    return 0.5 * forrester_high(x) + 10 * (x - 0.5) - 5


x_train_low = np.linspace(0, 1, 11)[:, None]
x_train_high = x_train_low[[0, 4, 6, 10]]
y_train_low = forrester_low(x_train_low)
y_train_high = forrester_high(x_train_high)

x_pred_low = np.linspace(0, 1, 100)[:, None]
x_pred_high = np.linspace(0, 1, 100)[:, None]

# MF_GPR
mf_gpr = pyLOM.GPR.MF_GPR(input_dim=1)
pyLOM.pprint(0,mf_gpr.kernel.available_kernels)  # Dict of available kernels
pyLOM.pprint(0,
    mf_gpr.kernel.get_kernel_parameters("Matern32")
)  # Dict of available parameters for the selecter kernel
kernel_low = mf_gpr.kernel.Matern32(
    variance=1.0,
    lengthscale=1.0,
)
kernel_high = mf_gpr.kernel.Matern32(
    variance=1.0,
    lengthscale=1.0,
)

mf_gpr.fit(
    train_features_list=[x_train_low, x_train_high],
    train_labels_list=[y_train_low, y_train_high],
    kernels=[kernel_low, kernel_high],
    noise_vars=[1e-6, 1e-4],
    num_restarts=7,
    verbose=True,
)
mf_gpr.display_model()
results = mf_gpr.predict([x_pred_low, x_pred_high])
mean_MF_LF = results["fidelity_1"]["mean"]
std_MF_LF = results["fidelity_1"]["std"]
mean_MF_HF = results["fidelity_2"]["mean"]
std_MF_HF = results["fidelity_2"]["std"]

plot_x = np.linspace(0, 1, 501).reshape(-1, 1)
plt.figure(figsize=(8,6),dpi=100)
plt.plot(plot_x, forrester_high(plot_x), linewidth=2, color="red", label="$True HF$")
plt.plot(
    plot_x,
    forrester_low(plot_x),
    linewidth=2,
    color="blue",
    linestyle="--",
    label="$True LF$",
)
plt.scatter(
    x_train_high,
    y_train_high,
    marker="o",
    facecolors="none",
    color="red",
    label="$Data HF$",
)
plt.scatter(
    x_train_low,
    y_train_low,
    marker="s",
    facecolors="none",
    color="blue",
    label="$Data LF$",
)
plt.plot(x_pred_low[:, 0], mean_MF_LF, linewidth=2, color="green", label="$Pred LF$")
plt.fill_between(
    x_pred_low[:, 0],
    (mean_MF_LF - 1.96 * std_MF_LF).flatten(),
    (mean_MF_LF + 1.96 * std_MF_LF).flatten(),
    facecolor="green",
    alpha=0.3,
    edgecolor="green",
    linewidth=1.5,
    label="LF 95% CI",
    zorder=5,
)
plt.plot(x_pred_high[:, 0], mean_MF_HF, linewidth=2, color="orange", label="$Pred HF$")
plt.fill_between(
    x_pred_high[:, 0],
    (mean_MF_HF - 1.96 * std_MF_HF).flatten(),
    (mean_MF_HF + 1.96 * std_MF_HF).flatten(),
    facecolor="orange",
    alpha=0.3,
    edgecolor="orange",
    linewidth=1.5,
    label="HF 95% CI",
    zorder=5,
)
plt.xlim([0, 1])
plt.ylim([-10, 20])
plt.xlabel("$x$")
plt.ylabel("$y$")
plt.legend(loc=2)
plt.tight_layout()

pyLOM.cr_info()
plt.show()


#!/usr/bin/env python
#
# Manifold learning methodologies.
#
# Last revision: 11/02/2025
import numpy as np
from scipy.sparse.csgraph import shortest_path
from scipy.linalg import eigh

from ..utils.gpu import cp
from ..vmmath    import euclidean_d
from ..utils     import cr_nvtx as cr, raiseError, pprint


@cr('MANIFOLD.isomap')
def isomap(X:np.ndarray, dims:int, n_size:int, comp:int = 1 ,verbose:bool = True):
    """
    Computes Isomap embedding using the algorithm of Tenenbaum, de Silva, and Langford (2000).
    
    Parameters:
    X : ndarray
        NxM Data matrix with N points in the mesh for M simulations
    dims: int
        Embedding dimensionality to use 
    n_size : int
        Neighborhood size (number of neighbors for 'k' method)
    comp: int
        Component to embed, if more than 1 (defaults to 1, the largest)
    verbose: bool
        Display information (default is True)
    
    Returns:
    Y : ndarray
        Contains coordinates for d-dimensional embeddings in Y.
    R : list
        Residual variances for the embedding in Y.
    E : ndarray
        Edge matrix for neighborhood graph.
    """    
    # Compute pairwise distances in a condensed form and convert to a square form
    D = cp.asnumpy(euclidean_d(X)) if type(X) is cp.ndarray else euclidean_d(X)

    # Step 0: Initialization and Parameters
    N = D.shape[0]
    if D.shape[1] != N:
        raise ValueError("D must be a square matrix")

    K = n_size
    if not isinstance(K, int) or K <= 0:
        raiseError("Number of neighbors for 'k' method must be a positive integer")
    
    INF = 1000 * np.max(D) * N  # Effectively infinite distance
        
    # Step 1: Construct neighborhood graph
    if verbose:
        pprint(0,"Constructing neighborhood graph...",flush=True)
        
    # Sort distances and keep only K-nearest neighbors
    sorted_indices = np.argsort(D, axis=1)
    for i in range(N):
        D[i, sorted_indices[i, K+1:]] = INF  # Only keep the K nearest neighbors

    D = np.minimum(D, D.T)  # Ensure distance matrix is symmetric

    E = (D != INF).astype(int)  # Edge matrix for neighborhood graph

    # Step 2: Compute shortest paths using Floyd-Warshall algorithm
    if verbose:
        pprint(0,"Computing shortest paths...",flush=True)

    G = shortest_path(D, method='FW', directed=False, unweighted=False)

    # Step 3: Identify connected components
    firsts = np.min((G == INF).astype(int) * np.arange(N), axis=1)
    comps, indices = np.unique(firsts, return_inverse=True)
    size_comps = np.array([np.sum(indices == i) for i in range(len(comps))])
    
    sorted_indices = np.argsort(size_comps)[::-1]
    comps = comps[sorted_indices]
    size_comps = size_comps[sorted_indices]

    if comp > len(comps):
        comp = 1  # Default to largest component if specified component is out of range

    if verbose:
        pprint(0,f"Number of connected components in graph: {len(comps)}",flush=True)
        pprint(0,f"Embedding component {comp} with {size_comps[comp-1]} points.",flush=True)

    # Select the largest connected component
    index = np.where(firsts == comps[comp-1])[0]
    G = G[np.ix_(index, index)]
    N = len(index)

    # Step 4: Construct low-dimensional embeddings using Classical MDS
    H = np.eye(N) - np.ones((N, N)) / N
    B = -0.5 * H @ (G**2) @ H  # Centering the matrix
    
    eigenvalues, eigenvectors = eigh(B, subset_by_index=[N - dims, N - 1])
    
    # Sorting eigenvalues and corresponding eigenvectors in descending order
    idx = np.argsort(eigenvalues)[::-1]
    eigenvalues = eigenvalues[idx]
    eigenvectors = eigenvectors[:, idx]

    # Compute embeddings for each specified dimension
    if dims <= N:
        coords = eigenvectors[:, :dims] * np.sqrt(eigenvalues[:dims])
        Y = coords.T
        
        # Compute residual variance
        L2_D = np.linalg.norm(coords[:, None] - coords[None, :], axis=2).flatten()
        r2 = 1 - np.corrcoef(L2_D, G.flatten())[0, 1]**2
        R = r2

        if verbose:
            pprint(0,f"Isomap on {N} points with dimensionality {dims} --> residual variance = {R}",flush=True)
    else:
        raiseError('Selected number of dimensions is higher than the number of samples')
        return None, None, None

    return Y, R, E


@cr('MANIFOLD.mds')
def mds(X:np.ndarray, dims:int, verbose:bool = True):
    """
    Computes the MDS embedding using a custom approach with squared distances and eigen-decomposition.
    
    Parameters:
    X : ndarray
        NxM Data matrix with N points in the mesh for M simulations
    dims : int
        Embedding dimensionality to use (p in your MATLAB code)
    
    Returns:
    Y : ndarray
        Contains coordinates for d-dimensional embeddings in Y.
    """
    # Step 1: Compute the pairwise Euclidean distance matrix and square it
    D = cp.asnumpy(euclidean_d(X)) if type(X) is cp.ndarray else euclidean_d(X)
    D = D * D

    # Step 2: Apply the custom centering formula to get matrix B
    n = D.shape[0]
    C = np.eye(n) - np.ones((n,n))/n
    B = -0.5*np.matmul(np.matmul(C,D),C)

    # Step 3: Compute eigen-decomposition
    # eigenvalues, eigenvectors = eigh((B + B.T) / 2)
    eigenvalues, eigenvectors = eigh(B)

    # Step 4: Sort eigenvalues and corresponding eigenvectors in descending order
    idx = np.argsort(eigenvalues)[::-1]
    eigenvalues = eigenvalues[idx]
    eigenvectors = eigenvectors[:, idx]
    eigenvalues = eigenvalues[:dims]
    eigenvectors = eigenvectors[:,:dims]

    # Step 5: Keep only positive eigenvalues beyond a roundoff threshold
    threshold = np.max(np.abs(eigenvalues)) * np.finfo(eigenvalues.dtype).eps**(3 / 4)
    keep = eigenvalues > threshold
    if not np.any(keep):
        Y = np.zeros((n, 1))
    else:
        # Compute the embedding with kept eigenvalues and corresponding eigenvectors
        Y = eigenvectors[:, keep] * np.sqrt(eigenvalues[keep])

    return Y.T


#!/usr/bin/env python
#
# PYLOM Testsuite 2D-VAE.
#
# Last revision: 24/09/2024
from __future__ import print_function, division

import mpi4py
mpi4py.rc.recv_mprobe = False

import sys, os, numpy as np
import pyLOM, pyLOM.NN

DATAFILE  = sys.argv[1]
VARIABLES = eval(sys.argv[2])
OUTDIR    = sys.argv[3]


## Set device
device = pyLOM.NN.select_device('cpu')


## Specify autoencoder parameters
nlayers     = 5
channels    = 64
lat_dim     = 5
beta        = 0
beta_start  = 0
beta_wmup   = 0
kernel_size = 4
nlinear     = 512
padding     = 1
activations = [pyLOM.NN.relu(), pyLOM.NN.relu(), pyLOM.NN.relu(), pyLOM.NN.relu(), pyLOM.NN.relu(), pyLOM.NN.relu(), pyLOM.NN.relu()]


## Load pyLOM dataset and set up results output
RESUDIR = os.path.join(OUTDIR,'vae_beta_%.2e_ld_%i' % (beta, lat_dim))
pyLOM.NN.create_results_folder(RESUDIR,verbose=False)


## Load pyLOM dataset
m    = pyLOM.Mesh.load(DATAFILE)
d    = pyLOM.Dataset.load(DATAFILE,ptable=m.partition_table)
u    = d.X(*VARIABLES)
um   = pyLOM.math.temporal_mean(u)
u_x  = pyLOM.math.subtract_mean(u, um)
time = d.get_variable('time')


## Mesh size
n0h = len(np.unique(pyLOM.utils.round(d.xyz[:,0],5)))
n0w = len(np.unique(pyLOM.utils.round(d.xyz[:,1],5)))
nh  = 448
nw  = 192


## Create a torch dataset
td   = pyLOM.NN.Dataset((u_x,), (n0h, n0w))
td.crop(nh, nw)


## Set and train the variational autoencoder
betasch    = pyLOM.NN.betaLinearScheduler(0., beta, beta_start, beta_wmup)
encoder    = pyLOM.NN.Encoder2D(nlayers, lat_dim, nh, nw, td.num_channels, channels, kernel_size, padding, activations, nlinear, vae=True)
decoder    = pyLOM.NN.Decoder2D(nlayers, lat_dim, nh, nw, td.num_channels, channels, kernel_size, padding, activations, nlinear)
model      = pyLOM.NN.VariationalAutoencoder(lat_dim, (nh, nw), td.num_channels, encoder, decoder, device=device)
early_stop = pyLOM.NN.EarlyStopper(patience=5, min_delta=0.02)

pipeline = pyLOM.NN.Pipeline(
    train_dataset = td,
    test_dataset  = td,
    model=model,
    training_params={
        "batch_size": 4,
        "epochs": 10,
        "lr": 1e-4,
        "betasch": betasch,
        "BASEDIR":RESUDIR
    },
)
pipeline.run()


## Reconstruct dataset and compute accuracy
rec = model.reconstruct(td)
rd  = pyLOM.NN.Dataset((rec), (nh, nw))
rd.pad(n0h, n0w)
td.pad(n0h, n0w)
d.add_field('urec',1,rd[:,0,:,:].numpy().reshape((len(time),n0w*n0h)).T)
d.add_field('utra',1,td[:,0,:,:].numpy().reshape((len(time),n0w*n0h)).T)
pyLOM.io.pv_writer(m,d,'reco',basedir=RESUDIR,instants=np.arange(time.shape[0],dtype=np.int32),times=time,vars=VARIABLES+['urec', 'utra'],fmt='vtkh5')


## Testsuite output
pyLOM.pprint(0,'TSUITE u_x  =',u_x.min(),u_x.max(),u_x.mean())
#pyLOM.pprint(0,'TSUITE urec =',d['urec'].min(),d['urec'].max(),d['urec'].mean())
pyLOM.pprint(0,'TSUITE utra =',d['utra'].min(),d['utra'].max(),d['utra'].mean())


pyLOM.cr_info()
pyLOM.pprint(0,'End of output')
